package com.yanzuoguang.util.thread;

import com.yanzuoguang.util.exception.CodeException;
import com.yanzuoguang.util.helper.StringHelper;

import java.util.Date;

/**
 * 自动执行队列的某一项
 */
public class RunnableListAutoItem implements Comparable<RunnableListAutoItem> {

    /**
     * 私有构造函数
     */
    private RunnableListAutoItem() {

    }

    /**
     * 构造函数
     *
     * @param code 需要执行的代码
     */
    public RunnableListAutoItem(Runnable code) {
        this("", code);
    }

    /**
     * 构造函数
     *
     * @param name
     * @param code
     */
    public RunnableListAutoItem(String name, Runnable code) {
        this();
        if (code == null) {
            return;
        }
        if (name == null) {
            name = "";
        }
        this.ClassType = code.getClass();
        this.Name = name;
        this.code = code;
        this.Key = StringHelper.getMD5Id(this.ClassType.getName() + ":" + name);
    }

    /**
     * 关键字
     */
    private volatile String Key;

    /**
     * 类型
     */
    private volatile Class<?> ClassType;

    /**
     * 名称
     */
    private volatile String Name;

    /**
     * 执行代码
     */
    private volatile Runnable code;

    /**
     * 总执行时间
     */
    private volatile long Time;

    /**
     * 执行次数
     */
    private volatile int Count;

    /**
     * 需要注意当达到 Time > Long.MaxValue / 2 或者 count > Integer.MaxValue /2 时
     * 则Time = Time/2, count  = count/2;
     */
    private synchronized void init() {
        if (this.Time > Long.MAX_VALUE / 2 || this.Count > Integer.MAX_VALUE / 2) {
            this.Time = this.Time / 2;
            this.Count = this.Count / 2;
        }
    }

    /**
     * 添加执行时间
     *
     * @param to
     */
    public synchronized void add(RunnableListAutoItem to) {
        if (this.ClassType != to.ClassType || !this.Name.equals(to.Name)) {
            return;
        }
        this.Time += to.Time;
        this.Count += to.Count;
        this.init();
    }

    /**
     * 获取平均时间
     *
     * @return
     */
    public synchronized long getAvgTime() {
        if (this.Count == 0) {
            return 0;
        }
        return this.Time / this.Count;
    }

    /**
     * 创建一个新的实体
     *
     * @return
     */
    public synchronized RunnableListAutoItem newBase() {
        RunnableListAutoItem to = new RunnableListAutoItem();
        to.Key = this.Key;
        to.ClassType = this.ClassType;
        to.Name = this.Name;
        to.Count = this.Count;
        to.Time = this.Time;
        return to;
    }

    /**
     * 开始执行函数
     */
    public synchronized void execute() {
        Date start = new Date();
        try {
            if (this.code == null) {
                return;
            }
            this.code.run();
        } catch (Exception ex) {
            this.HandleException(ex);
        } finally {
            this.Count++;
            this.Time += new Date().getTime() - start.getTime();
        }
    }

    /**
     * 处理异常数据
     *
     * @param ex
     */
    private void HandleException(Throwable ex) {
        if (ex instanceof RuntimeException) {
            throw (RuntimeException) ex;
        } else {
            throw new CodeException(ex.getMessage(), ex);
        }
    }

    @Override
    public int compareTo(RunnableListAutoItem to) {
        return (int) (this.getAvgTime() - to.getAvgTime());
    }

    public String getKey() {
        return Key;
    }

    public Class<?> getClassType() {
        return ClassType;
    }

    public String getName() {
        return Name;
    }

    public Runnable getCode() {
        return code;
    }

    public long getTime() {
        return Time;
    }

    public int getCount() {
        return Count;
    }
}
