package com.yanzuoguang.cloud.aop;

import com.alibaba.fastjson.JSON;
import com.yanzuoguang.util.cache.MemoryCache;
import com.yanzuoguang.util.contants.ResultConstants;
import com.yanzuoguang.util.exception.ExceptionHelper;
import com.yanzuoguang.util.helper.JSONHelper;
import com.yanzuoguang.util.helper.StringHelper;
import com.yanzuoguang.util.log.Log;
import com.yanzuoguang.util.base.ObjectHelper;
import com.yanzuoguang.util.thread.ThreadNext;
import com.yanzuoguang.util.vo.LogVo;
import com.yanzuoguang.util.vo.ResponseResult;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.Signature;
import org.aspectj.lang.annotation.Around;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.annotation.Pointcut;
import org.aspectj.lang.reflect.MethodSignature;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.ApplicationContext;
import org.springframework.stereotype.Component;

import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.lang.reflect.Method;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.LinkedBlockingQueue;

/**
 * LogsAspect（接口请求日志切面）
 *
 * @author: Kang
 * @time: 2018年04月25日 11:43
 */
@Aspect
@Component
public class WebAspect extends AbstractValidateAspect implements ThreadNext.Next {

    private static final Logger logger = LoggerFactory.getLogger(WebAspect.class);

    @Value("${spring.application.name}")
    private String applicationName;

    @Value("${yzg.logAll:false}")
    private boolean logAll = false;

    @Value("${yzg.cacheTime:120}")
    private int cacheTime = 120;

    @Autowired
    private ApplicationContext context;

    private LogFeign logFeign;

    /**
     * 缓存队列
     */
    private volatile LinkedBlockingQueue<LogVo> cache = new LinkedBlockingQueue<LogVo>();

    // 缓存的每次请求的结果
    private MemoryCache<RequestCacheResult> _cacheResult = new MemoryCache<>(cacheTime);

    public WebAspect() {
        ThreadNext.start(this, "save log error");
    }

    /**
     * exec aop point aspect
     */
    @Pointcut("execution(* *..web..*Controller.*(..))")
    public void webAspect() {
    }

    /**
     * 执行环形切面
     *
     * @param joinPoint
     * @return
     */
    @Around(value = "webAspect()")
    public Object requestWebAround(ProceedingJoinPoint joinPoint) throws Throwable {
        Log.threadBegin();
        ResponseResult responseResult = null;           // 用户数据库记录
        long start = System.currentTimeMillis();
        String name = joinPoint.getSignature().getName();
        try {
            logger.info("[ {} ] request params is {}", name, joinPoint.getArgs());

            Object result = executeMethod(joinPoint, name);
            if (result instanceof ResponseResult) {
                responseResult = (ResponseResult) result;
            } else {
                responseResult = ResponseResult.result(result);
            }

            long end = System.currentTimeMillis();
            logger.info("[ {} ] time ({})ms, result is {}", name, (end - start), responseResult);

            return result;
        } catch (Exception e) {
            long end = System.currentTimeMillis();
            logger.error("[ {} ] time ({})ms,  is error {}", name, (end - start), e);
            responseResult = ExceptionHelper.getError(e);
            if (getReturnType(joinPoint).getTypeName().indexOf(ResponseResult.class.getName()) > -1) {
                return responseResult;
            } else {
                throw e;
            }
        } finally {
            Log.threadCommit();
            saveInterLogs(joinPoint, responseResult);
        }
    }

    /**
     * 获取返回的至类型
     *
     * @param joinPoint
     * @return
     * @throws NoSuchMethodException
     */
    private Type getReturnType(ProceedingJoinPoint joinPoint) {
        //获取返回值类型
        Signature s = joinPoint.getSignature();
        MethodSignature ms = (MethodSignature) s;
        Method m = ms.getMethod();
        Type t = m.getAnnotatedReturnType().getType();
        return t;
    }

    private Object executeMethod(ProceedingJoinPoint joinPoint, String name) throws Throwable {
        if (joinPoint.getArgs().length != 1
                || joinPoint.getArgs().length == 1 &&
                (joinPoint.getArgs()[0] instanceof ServletResponse || joinPoint.getArgs()[0] instanceof ServletRequest)
                ) {
            return joinPoint.proceed();
        } else {
            // 获取请求编号
            Object firstArgs = joinPoint.getArgs().length > 0 ? joinPoint.getArgs()[0] : null;
            String reqId = ObjectHelper.getString(firstArgs, "reqId");
            if (StringHelper.isEmpty(reqId)) {
                reqId = StringHelper.md5(JSONHelper.serialize(firstArgs));
            } else {
                // 请求编号和公司编号挂钩
                reqId = StringHelper.getId(ObjectHelper.getString(firstArgs, "companyId"), reqId);
            }
            String reqFull = StringHelper.getId(reqId, getHttpRequestUrl());
            RequestCacheResult req = _cacheResult.get(reqFull, new RequestCacheResult(reqFull));
            // 缓存的键值对
            if (req.getResult() == null) {
                synchronized (req) {
                    if (req.getResult() == null) {
                        try {
                            req.setResult(joinPoint.proceed());
                        } catch (Exception ex) {
                            req.setResult(ex);
                        }
                    }
                }
            }
            Object result = req.getResult();
            if (result instanceof Throwable) {
                throw (Throwable) result;
            }
            return result;
        }
    }

    /**
     * deal with inter api logs
     *
     * @param joinPoint      请求参数
     * @param responseResult 返回参数
     */
    private void saveInterLogs(ProceedingJoinPoint joinPoint, ResponseResult responseResult) {
        // 日志请求不记录,防止死循环递归
        if (applicationName.indexOf("log") >= 0) {
            return;
        }
        // 正常请求不记录
        if (!logAll && responseResult.getCode() == ResultConstants.SUCCESS) {
            return;
        }
        LogVo logVo = initLogInterVo(getHttpRequestUrl(), joinPoint, responseResult);
        addLog(logVo);
    }

    /**
     * 添加日志到缓存中，并不是立即添加，而是通过线程执行，防止对环境造成影响
     *
     * @param logVo
     */
    public void addLog(LogVo logVo) {
        cache.add(logVo);
    }

    /**
     * 初始化日志Vo
     *
     * @param url            请求路径
     * @param joinPoint      请求参数
     * @param responseResult 返回参数
     * @return
     */
    private LogVo initLogInterVo(String url, ProceedingJoinPoint joinPoint, ResponseResult responseResult) {
        LogVo logInterVo = new LogVo();
        logInterVo.setLogId(StringHelper.getNewID());
        logInterVo.setLogSources(applicationName);//平台名
        logInterVo.setInterUrl(url);//请求URL
        List<Object> para = new ArrayList<>();
        for (Object item : joinPoint.getArgs()) {
            if (item instanceof HttpServlet || item instanceof HttpServletResponse || item instanceof HttpServletRequest) {
                continue;
            }
            para.add(item);
        }
        Object paraTo = para;
        if (para.size() == 1) {
            paraTo = para.get(0);
        }
        logInterVo.setContent(JSON.toJSONString(paraTo));//请求参数
        logInterVo.setResult(JSON.toJSONString(responseResult));//返回参数
        logInterVo.setStatus(responseResult.getCode() == ResultConstants.SUCCESS ? 1 : 0);
        return logInterVo;
    }

    /**
     * 执行下一个函数,出现异常会继续执行
     *
     * @return 是否继续执行
     */
    @Override
    public boolean next() {
        while (cache.size() > 0) {
            LogVo item = cache.poll();
            if (item != null) {
                if (logFeign == null) {
                    logFeign = context.getBean(LogFeign.class);
                }
                logFeign.save(item);
            }
        }
        return true;
    }

    /**
     * 沉睡时间
     *
     * @return
     */
    @Override
    public int getNextTime() {
        return 1000;
    }
}
